#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
from DbConnection import DatabaseConnection
from adapt_config import AdaptConfig
from misc_const import SqlConnectMode, sql_connect_mode
import os
import sys
import re
import traceback

InstanceOfThis = None


class DatabaseCalls():
    @staticmethod
    def GetOrCreateInstance(database_hostname, database_port, database_user, database_pass, database_dbname):
        global InstanceOfThis
        if(InstanceOfThis is None):
            InstanceOfThis = DatabaseCalls(database_hostname, database_port, database_user, database_pass, database_dbname)
            
        return InstanceOfThis
    
    def Clone(self):
        
        return DatabaseCalls(self.Hostname, self.Port, self.User, self.Password, self.dbname)
    
    # Routine here to enable direct queries to allow system to work during refactoring.
    # Eventually this routine will be replaced. All writes should go through a member
    # function created for a specific purpose.
    def ExecuteWriteQuery(self, Query, Val=None):
        self._SqlDataProvider.ExecuteWriteQuery(Query, Val=None)
    
    def ExecuteConnectionCommit(self):
        self._SqlDataProvider.ExecuteConnectionCommit()
    
    def ExplicitlyClose(self):
        self._SqlDataProvider.ExplicitlyClose()
        del self._SqlDataProvider
    
    def __init__(self, database_hostname, database_port, database_user, database_password, database_name):
        self.Hostname = database_hostname
        self.Port = database_port
        self.User = database_user
        self.Password = database_password
        self.dbname = database_name
        
        self._SqlDataProvider = DatabaseConnection(database_hostname, database_port, database_user, database_password, database_name)
        
    def __del__(self):
        # Intentionally NOT closing database connection. This will be done when DatabaseConnection goes out of scope.
        pass
    
    def ExecuteLegacyReadQuery(self, SqlQuery):
        return self._SqlDataProvider.ExecuteLegacyReadQuery(SqlQuery)
    
    def experimentsSelectAllAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT id, name, description, starttime, probability_threshold, is_checkpointed, retries_permitted FROM experiments ORDER BY id;""")
    
    def experimentsSelectAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, name, description, starttime, probability_threshold, is_checkpointed, retries_permitted FROM experiments WHERE id = %d;""" % (int(db_expid)))
    
    def experimentsSelectLocalStorageDirWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT localstoragedir FROM experiments WHERE id = %d;""" % (int(db_expid)))
    
    def experimentsSelectRetriesPermittedWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT retries_permitted FROM experiments WHERE id = %d;""" % (int(db_expid)))
    
    def experimentsSelectIdAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT id FROM experiments;""")
    
    def experimentsInsertAllAll(self, **kwargs):
        db_name = kwargs.get('name', '')
        db_description = kwargs.get('description', '')
        db_probability_threshold = kwargs.get('probability_threshold', '')
        db_checkpointscript = kwargs.get('checkpointscript', '')
        db_localstoragedir = kwargs.get('localstoragedir', '')
        db_y_or_n = kwargs.get('y_or_n', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO experiments (name, description, starttime, probability_threshold, checkpointscript, localstoragedir, is_checkpointed, retries_permitted) VALUES ('%s', '%s', now(), %d, '%s',%s, '%s', 5);""" % (db_name, db_description, float(db_probability_threshold), db_checkpointscript, db_localstoragedir, db_y_or_n))
        return
    
    def experimentsUpdateCheckpointStatusWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_y_or_n = kwargs.get('y_or_n', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE experiments SET is_checkpointed = '%s' WHERE id = %d;""" % (db_y_or_n, int(db_expid)))
        self._SqlDataProvider.ExecuteConnectionCommit()
        return
    
    def experimentsUpdateProbabilityThresholdWhere(self, **kwargs):
        db_probability_threshold = kwargs.get('probability_threshold', '')
        db_expid = kwargs.get('expid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE experiments SET probability_threshold = %d WHERE id = %d;""" % (float(db_probability_threshold), int(db_expid)))
        return
    
    def experimentsUpdateRetriesPermittedWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_retries_permitted = kwargs.get('retries_permitted', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE experiments SET retries_permitted = %d WHERE id = %d;""" % (int(db_retries_permitted), int(db_expid)))
        return
    
    def experimentsDeleteAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM experiments WHERE id = %d;""" % (int(db_expid)))
        return
    
    def control_messagesInsertServerStopAll(self):
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO control_messages VALUES ('server_stop','1');""")
        self._SqlDataProvider.ExecuteConnectionCommit()
        return
    
    def control_messagesDeleteServerStopAll(self):
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM control_messages WHERE k = 'server_stop';""")
        return
    
    def control_messagesSelectServerStopAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT v FROM control_messages WHERE k = 'server_stop';""")
    
    def simulators_prereqsInsertInstallerAll(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        db_filename = kwargs.get('filename', '')
        db_inputtype = kwargs.get('inputtype', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO simulators_prereqs (simname, filename, inputtype) VALUES (%s, %s, %s);""", (db_simname, db_filename, db_inputtype))
        return
    
    def simulators_prereqsSelectFilenameWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        return self.ExecuteLegacyReadQuery("""SELECT filename FROM simulators_prereqs WHERE simname = '%s' AND inputtype = 'user';""" % (db_simname))
    
    def simulators_prereqsDeleteAllWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM simulators_prereqs WHERE simname = '%s';""" % (db_simname))
        return
    
    def simulators_installer_prereqsInsertAllAll(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        db_filename = kwargs.get('filename', '')
        db_payload = kwargs.get('payload', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO simulators_installer_prereqs (simname, filename, payload) VALUES (%s, %s, %s);""", (db_simname, db_filename, db_payload))
        return
    
    def simulators_installer_prereqsSelectFilenamePayloadWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        return self.ExecuteLegacyReadQuery("""SELECT filename, payload FROM simulators_installer_prereqs WHERE simname = '%s';""" % (db_simname))
    
    def simulators_installer_prereqsSelectFilenameWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        return self.ExecuteLegacyReadQuery("""SELECT filename FROM simulators_installer_prereqs WHERE simname = '%s';""" % (db_simname))
    
    def simulators_installer_prereqsUpdatePayloadAll(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        db_filename = kwargs.get('filename', '')
        db_payload = kwargs.get('payload', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE simulators_installer_prereqs SET payload = %s WHERE simname = %s AND filename = %s;""", (db_payload, str(db_simname), str(db_filename)))
        return
    
    def simulators_installer_prereqsDeleteAllWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM simulators_installer_prereqs WHERE simname = '%s';""" % (db_simname))
        return
    
    def simulatorsSelectAllWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        return self.ExecuteLegacyReadQuery("""SELECT simname, webwrapperscript, mainwrapperscript FROM simulators WHERE simname = '%s';""" % (db_simname))
    
    def simulatorsSelectAllAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT simname, webwrapperscript, mainwrapperscript FROM simulators;""")
    
    def simulatorsInsertAllAll(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        db_webwrappertxt = kwargs.get('webwrappertxt', '')
        db_mainwrappertxt = kwargs.get('mainwrappertxt', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO simulators (simname, webwrapperscript, mainwrapperscript) VALUES (%s, %s, %s);""", (db_simname, db_webwrappertxt, db_mainwrappertxt))
        return
    
    def simulatorsDeleteAllWhere(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM simulators WHERE simname = '%s';""" % (db_simname))
        self._SqlDataProvider.ExecuteConnectionCommit()
        return
    
    def simulatorsUpdateAllWhereSimname(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        db_value = kwargs.get('value', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE simulators SET %s WHERE simname = '%s';""" % (db_value, db_simname))
        return
    
    def branch_input_blobsDeleteAllWhereIn(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branch_input_blobs WHERE branchid IN (%s);""" % (db_relevant_branches_set))
        return
    
    def branch_input_blobsDeleteAllWhere(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branch_input_blobs WHERE branchid = %d;""" % (int(db_branchid)))
        return
    
    def branch_input_blobsSelectAllWhere(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, filename, permissions, dat FROM branch_input_blobs WHERE branchid = %d;""" % (int(db_branchid)))
        
    def branch_input_blobsSelectAllWhereInBranchID(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, filename, permissions, dat FROM branch_input_blobs WHERE branchid IN (%s);""" % (db_relevant_branches_set))
    
    def branch_input_blobsSelectAllWhereInBranchIDandState(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, filename, permissions, dat FROM branch_input_blobs WHERE branchid IN (%s) AND filename LIKE '%%.state';""" % (db_relevant_branches_set))
    
    def branch_input_blobsSelectAllWhereInExpIDandState(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, filename, permissions, dat FROM branch_input_blobs WHERE branchid IN (SELECT id FROM branches WHERE experimentid='%s') AND filename LIKE '%%.state';""" % (db_expid))
    
    def branch_input_blobsSelectAllWhereBranchIDAndFilename(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, dat FROM branch_input_blobs WHERE (branchid='%s' AND filename LIKE '%%.state');""" % (int(db_branchid)))
    
    def branch_input_blobsInsert(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_filename = kwargs.get('filename', '')
        db_permissions = kwargs.get('permissions', '')
        db_dat = kwargs.get('dat', '')
        db_dat = self.db_escape(db_dat)
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO branch_input_blobs(branchid, filename, permissions, dat) VALUES (%d, "%s", %d, "%s");""" % (int(db_branchid), db_filename, int(db_permissions), db_dat))
        return
    
    def branch_input_refsDeleteAllWhereIn(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branch_input_refs WHERE branchid IN (%s);""" % (db_relevant_branches_set))
        return
    
    def branch_input_refsDeleteAllWhere(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branch_input_refs WHERE branchid = %d;""" % (int(db_branchid)))
        return
    
    def branch_input_refsSelectAllWhere(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid, hostname, filenamehere, filenamethere FROM branch_input_refs WHERE branchid = %d;""" % (int(db_branchid)))
    
    def branching_condition_namesSelectAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT expid, code, name FROM branching_condition_names WHERE expid = %d;""" % (int(db_expid)))
    
    def branching_condition_namesInsert(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_code = kwargs.get('code', '')
        db_name = kwargs.get('name', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO branching_condition_names(expid, code, name) VALUES (%d, "%s", "%s");""" % (int(db_expid), str(db_code), str(db_name)))
        return
    
    def dynamic_importance_measuresInsertAllType12(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_bc_chosen = kwargs.get('bc_chosen', '')
        db_var_chosen = kwargs.get('var_chosen', '')
        db_DYI_name = kwargs.get('DYI_name', '')
        db_DYI_type = kwargs.get('DYI_type', '')
        db_consequence_file = kwargs.get('consequence_file', '')
        db_var_value_nonoccurrence = kwargs.get('var_value_nonoccurrence', '')
        db_value = kwargs.get('value', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO dynamic_importance_measures (expid, code, variable, name, type, consequence_file, value_nonoccurrence, value) VALUES (%d, "%s", "%s", "%s", %d, "%s", %s, "%s");""" % (int(db_expid), db_bc_chosen, db_var_chosen, db_DYI_name, int(db_DYI_type), db_consequence_file, db_var_value_nonoccurrence, db_value))
        return
    
    def dynamic_importance_measuresInsertAllType3(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_bc_chosen = kwargs.get('bc_chosen', '')
        db_var_chosen = kwargs.get('var_chosen', '')
        db_DYI_name = kwargs.get('DYI_name', '')
        db_DYI_type = kwargs.get('DYI_type', '')
        db_consequence_file = kwargs.get('consequence_file', '')
        db_value = kwargs.get('value', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO dynamic_importance_measures (expid, code, variable, name, type, consequence_file, value) VALUES (%d, "%s", "%s", "%s", %d, "%s", "%s");""" % (int(db_expid), db_bc_chosen, db_var_chosen, db_DYI_name, int(db_DYI_type), db_consequence_file, db_value))
        return
    
    def job_output_blobsDeleteAllWhereIn(self, **kwargs):
        db_relevant_jobs_set = kwargs.get('relevant_jobs_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_blobs WHERE jobid IN (%s);""" % (db_relevant_jobs_set))
        return
    
    def job_output_blobsDeleteAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_blobs WHERE jobid = %d;""" % (db_jobid))
        return
    
    def job_output_infoDeleteAllWhereIn(self, **kwargs):
        db_relevant_jobs_set = kwargs.get('relevant_jobs_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_info WHERE jobid IN (%s);""" % (db_relevant_jobs_set))
        return
    
    def job_output_infoDeleteAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_info WHERE jobid = %d;""" % (db_jobid))
        return
    
    def job_output_infoSelectInfoWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT info.k, info.v FROM job_output_info info, jobs j, branches b WHERE info.jobid = j.id AND b.id = %d AND b.id = j.branchid;""" % (int(db_branchid)))
    
    def job_output_infoSelectInfoWhereKAndExpID(self, **kwargs):
        db_k = kwargs.get('k', '')
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT jobid, k, v FROM job_output_info WHERE ((k = '%s') AND (jobid = ANY(SELECT jobid FROM branches_completed WHERE experimentid = %d)));""" % (str(db_k), int(db_expid)))
    
    def job_output_infoSelectInfoWhereFinishedByExpID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT jobid, k, v FROM job_output_info WHERE (jobid = ANY(SELECT jobid FROM branches_completed WHERE experimentid = %d));""" % (int(db_expid)))
    
    def job_output_infoSelectInfoWhereJobAndCode(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT v FROM job_output_info WHERE k = 'code' AND jobid = %d;""" % (int(db_jobid)))
    
    def job_output_infoSelectEndStateTimes(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT jobid, v FROM job_output_info WHERE jobid IN (SELECT id FROM jobs WHERE branchid IN (SELECT id FROM branches WHERE id NOT IN (SELECT DISTINCT parentid FROM branches WHERE parentid IS NOT NULL) AND experimentid=%d)) AND k='simtime';""" % (int(db_expid)))
    
    def job_output_infoSelectInfoWhereJob(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT k, v FROM job_output_info WHERE jobid = %d;""" % (int(db_jobid)))
    
    def job_output_infoInsertAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        db_k = kwargs.get('k', '')
        db_v = kwargs.get('v', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO job_output_info (jobid, k, v) VALUES (%d, '%s', '%s');""" % (int(db_jobid), db_k, db_v))
        self._SqlDataProvider.ExecuteConnectionCommit()
        return
    
    def job_checkpoints_takenDeleteAllWhereIn(self, **kwargs):
        db_relevant_jobs_set = kwargs.get('relevant_jobs_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_checkpoints_taken WHERE jobid IN (%s);""" % (db_relevant_jobs_set))
        return
    
    def job_checkpoints_takenDeleteAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_checkpoints_taken WHERE jobid = %d;""" % (db_jobid))
        return
    
    def job_checkpoints_takenInsertCheckpointWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO job_checkpoints_taken (jobid, checkpoint_time) VALUES (%d, NOW());""" % (int(db_jobid)))
        return
    
    def job_checkpoints_takenSelectCheckpointTimeWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT SUM(unix_timestamp(resume_time) - unix_timestamp(checkpoint_time)) FROM job_checkpoints_taken WHERE resume_time IS NOT null AND jobid = %d;""" % (int(db_jobid)))
    
    def job_output_lsDeleteAllWhereIn(self, **kwargs):
        db_relevant_jobs_set = kwargs.get('relevant_jobs_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_ls WHERE jobid IN (%s);""" % (db_relevant_jobs_set))
        return
    
    def job_output_lsDeleteAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_ls WHERE jobid = %d;""" % (db_jobid))
        return
    
    def job_output_lsSelectPathWhereJobID(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT path, size FROM job_output_ls WHERE jobid= %d ORDER BY path;""" % (int(db_jobid)))
    
    def job_output_lsInsertAllAll(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        db_path = kwargs.get('path', '')
        db_size = kwargs.get('size', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO job_output_ls (jobid, path, size) VALUES (%s, '%s', %d);""" % (int(db_jobid), db_path, int(db_size)))
        return
    
    def jobsDeleteAllWhereIn(self, **kwargs):
        db_relevant_jobs_set = kwargs.get('relevant_jobs_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM jobs WHERE id IN (%s);""" % (db_relevant_jobs_set))
        return
    
    def jobsDeleteAllWhere(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM job_output_ls WHERE id = %d;""" % (db_jobid))
        return
    
    def jobsUpdateStateWhereID(self, **kwargs):
        db_state = kwargs.get('state', '')
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET state = '%s' WHERE id = %d;""" % (db_state, int(db_jobid)))
        return
    
    def jobsUpdateStateEndtimeWhereID(self, **kwargs):
        db_state = kwargs.get('state', '')
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET state = '%s', endtime=NOW() WHERE id = %d;""" % (db_state, int(db_jobid)))
        return
    
    def jobsUpdateStatePIDWhereID(self, **kwargs):
        db_state = kwargs.get('state', '')
        db_branchid = kwargs.get('branchid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET state = '%s', pid = null WHERE branchid = %d;""" % (db_state, int(db_branchid)))
        return
    
    def jobsUpdateStatePIDWhereinID(self, **kwargs):
        db_jobids_aborted = kwargs.get('jobids_aborted', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET state = 'A', pid = null WHERE id IN (%s);""" % (db_jobids_aborted))
        return
    
    def jobsUpdateStateWhereState(self, **kwargs):
        db_state = kwargs.get('state', '')
        db_newstate = kwargs.get('newstate', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET state = '%s' WHERE state = '%s';""" % (db_newstate, db_state))
        return
    
    def jobsUpdateHostnameWhereID(self, **kwargs):
        db_hostname = kwargs.get('hostname', '')
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET hostname = '%s' WHERE id = %d;""" % (db_hostname, int(db_jobid)))
        return
    
    def jobsUpdateRundirWhereID(self, **kwargs):
        db_jobdir = kwargs.get('jobdir', '')
        db_jobid = kwargs.get('jobid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE jobs SET rundir = '%s' WHERE id = %d;""" % (db_jobdir, int(db_jobid)))
        return
    
    def jobsInsertAllAll(self, branchid, hostname):
        db_branchid = branchid
        db_hostname = hostname
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO jobs (branchid, state, hostname, starttime) VALUES (%d, 'R', '%s', now());""" % (int(db_branchid), db_hostname))
        return
    
    def jobsInsertAllReduced(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_state = kwargs.get('state', '')
        db_rundir = kwargs.get('rundir', '')
        db_hostname = kwargs.get('hostname', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO jobs (branchid, state, hostname, rundir, starttime) VALUES (%d, '%s', '%s', '%s', now());""" % (int(db_branchid), str(db_state), str(db_hostname), str(db_rundir)))
        return
    
    def jobsSelectAllWhereInBranchID(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, state, hostname, rundir, pid FROM jobs WHERE branchid IN (%s);""" % (db_relevant_branches_set))
    
    def jobsSelectAllWhereInExpID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT * FROM jobs WHERE branchid IN (SELECT id FROM branches WHERE experimentid=%d);""" % (int(db_expid)))
    
    def jobsSelectAllWhereStateAndBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_state = kwargs.get('state', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, branchid, state, hostname, rundir, starttime, endtime, pid FROM jobs WHERE branchid = %d and state='%s';""" % (int(db_branchid), db_state))
    
    def jobsSelectAllWhereStateAndJobID(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        db_state = kwargs.get('state', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, branchid, state, hostname, rundir, starttime, endtime, pid FROM jobs WHERE id = %d and state='%s';""" % (int(db_jobid), db_state))
    
    def jobsSelectAllWhereState(self, **kwargs):
        db_state = kwargs.get('state', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, branchid, state, hostname, rundir, starttime, endtime, pid FROM jobs WHERE state='%s';""" % (db_state))
    
    def LEGACYjobsSelectIDWhereBranchIDAndState(self, db_branchid):
        SqlQuery = """SELECT id FROM jobs WHERE branchid = %d AND (state = 'C' OR %d IN (SELECT branchid FROM branches_persistently_failed));""" % (int(db_branchid), int(db_branchid))
        return self._SqlDataProvider.ExecuteLegacyReadQuery(SqlQuery)
    
    def jobsSelectIDWhereBranchIDAndState(self, db_branchid):
        SqlQuery = """SELECT id FROM jobs WHERE branchid = %d AND (state = 'C' OR %d IN (SELECT branchid FROM branches_persistently_failed));""" % (int(db_branchid), int(db_branchid))
        return self.dbreadcursor.execute(SqlQuery)
    
    def jobsSelectAllWhereBranch(self, branchid):
        return self._SqlDataProvider.ExecuteReadQuery("""SELECT * from jobs WHERE branchid = %d;""" % (int(branchid)))
    
    def jobsSelectAllWhereJobID(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, branchid, state, hostname, rundir, starttime, endtime, pid FROM jobs WHERE id = %d;""" % (int(db_jobid)))
    
    def jobsSelectAllWhereExpIDCompleted(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, branchid, state, hostname, rundir, starttime, endtime, pid FROM jobs WHERE (id = ANY(SELECT jobid FROM branches_completed WHERE experimentid = %d));""" % (int(db_expid)))
    
    def jobsSelectHostnameWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT hostname FROM jobs WHERE branchid = %d;""" % (int(db_branchid)))
    
    def jobsSelectHostnameWhereJobID(self, **kwargs):
        db_jobid = kwargs.get('jobid', '')
        return self.ExecuteLegacyReadQuery("""SELECT hostname, rundir FROM jobs WHERE id = %d;""" % (int(db_jobid)))
    
    def jobsSelectHostnameWhereState(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_state = kwargs.get('state', '')
        return self.ExecuteLegacyReadQuery("""SELECT hostname FROM jobs WHERE branchid = %d AND state = '%s';""" % (int(db_branchid), db_state))
    
    def branchesDeleteAllWhereIn(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branches WHERE id IN (%s);""" % (db_relevant_branches_set))
        return
    
    def branchesDeleteAllWhere(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM branches WHERE id = %d;""" % (int(db_branchid)))
        return
    
    def branchesSelectIDWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE id = %d ORDER BY id DESC;""" % (int(db_branchid)))
    
    def branchesSelectIDWhereParentID(self, **kwargs):
        db_parentid = kwargs.get('parentid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE parentid = %d ORDER BY id DESC;""" % (int(db_parentid)))
    
    def branchesSelectIDWhereInParentIDAndNotID(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE parentid IN (%s) AND id NOT IN (%s);""" % (db_relevant_branches_set, db_relevant_branches_set))
    
    def branchesSelectIDWhereExpIDAndParentIDandDescription(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_parentid = kwargs.get('parentid', '')
        db_description = kwargs.get('description', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE (experimentid = %d AND parentid = %d AND branchdescription = '%s');""" % (int(db_expid), int(db_parentid), db_description))
    
    def branchesSelectEndStates(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE id NOT IN (SELECT DISTINCT parentid FROM branches WHERE parentid IS NOT NULL) AND experimentid=%d;""" % (int(db_expid)))
    
    def branchesSelectAllInExp(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id FROM branches WHERE experimentid=%d;""" % (int(db_expid)))
    
    def branchesSelectEndStatesProbability(self, **kwargs):
        db_experimentid = kwargs.get('experimentid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, probability FROM branches WHERE id NOT IN (SELECT DISTINCT parentid FROM branches WHERE parentid IS NOT NULL) AND experimentid=%d;""" % (int(db_experimentid)))
    
    def branchesSelectParentIDWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT parentid FROM branches WHERE id = %d;""" % (int(db_branchid)))
    
    def branchesSelectParentIDWhereExpID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT parentid FROM branches WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branchesSelectIDParentIDWhereExpID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, parentid FROM branches WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branchesSelectAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, experimentid, parentid, branchdescription, jobscript, submittime, probability, probability_threshold_multiplier, terminate_early, priority, executable FROM branches WHERE experimentid = %d ORDER BY id ASC;""" % (int(db_expid)))
    
    def branchesSelectSomeWhereExpid(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, experimentid, parentid, branchdescription, probability FROM branches WHERE experimentid = %d ORDER BY id ASC;""" % (int(db_expid)))
    
    def branchesSelectAllWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, experimentid, parentid, branchdescription, jobscript, submittime, probability, probability_threshold_multiplier, terminate_early, priority, executable FROM branches WHERE id = %d;""" % (int(db_branchid)))
    
    def branchesSelectProbWhereInBranchID(self, **kwargs):
        db_relevant_branches_set = kwargs.get('relevant_branches_set', '')
        return self.ExecuteLegacyReadQuery("""SELECT id, probability FROM branches WHERE id in (%s);""" % (db_relevant_branches_set))
    
    def branchesInsertAllRoot(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_description = kwargs.get('description', '')
        db_jobscript = kwargs.get('jobscript', '')
        db_probability = kwargs.get('probability', '')
        db_probability_threshold_multiplier = kwargs.get('probability_threshold_multiplier', '')
        db_terminate_early = kwargs.get('terminate_early', '')
        db_priority = kwargs.get('priority', '')
        db_branchexecutable = kwargs.get('branchexecutable', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO branches (experimentid, branchdescription, jobscript, submittime, probability, probability_threshold_multiplier, terminate_early, priority, executable) VALUES (%d, '%s', '%s', NOW(), %d, %d, %d, %d, '%s');""" % (int(db_expid), db_description, db_jobscript, float(db_probability), float(db_probability_threshold_multiplier), int(db_terminate_early), int(db_priority), db_branchexecutable))
        return
    
    def branchesInsertAllChild(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_parentid = kwargs.get('parentid', '')
        db_description = kwargs.get('description', '')
        db_probability = kwargs.get('probability', '')
        db_probability_threshold_multiplier = kwargs.get('probability_threshold_multiplier', '')
        db_terminate_early = kwargs.get('terminate_early', '')
        db_priority = kwargs.get('priority', '')
        db_branchexecutable = kwargs.get('branchexecutable', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO branches (experimentid, parentid, branchdescription, submittime, probability, probability_threshold_multiplier, terminate_early, priority, executable) VALUES (%d, %d, '%s', NOW(), %s, %d, %d, %d, '%s');""" % (int(db_expid), int(db_parentid), str(db_description), db_probability, float(db_probability_threshold_multiplier), int(db_terminate_early), int(db_priority), str(db_branchexecutable)))
        return
    
    def branchesUpdatePriorityWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_priority = kwargs.get('priority', '')
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE branches SET priority = %d WHERE id = %d;""" % (int(db_priority), int(db_branchid)))
        return
    
    def branchesUpdateProbabilityWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        db_probability = kwargs.get('probability', '')
        self._SqlDataProvider.ExecuteConnectionCommit()
        self._SqlDataProvider.ExecuteWriteQuery("""UPDATE branches SET probability = '%s' WHERE id = %d;""" % (db_probability, int(db_branchid)))
        self._SqlDataProvider.ExecuteConnectionCommit()
        return
    
    def branchesSelectCountWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT COUNT(id) FROM branches WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branchesSelectDistinctParentIDWhereExperimentID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT distinct parentid FROM branches WHERE experimentid = %d;""" % (int(db_expid)))
    
    def experiment_environment_variablesDeleteAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM experiment_environment_variables WHERE experimentid = %d;""" % (int(db_expid)))
        return
    
    def experiment_environment_variablesSelectAllWhere(self, db_expid):
        return self.ExecuteLegacyReadQuery("""SELECT experimentid, k, v FROM experiment_environment_variables WHERE experimentid = %d;""" % (int(db_expid)))
    
    def experiment_environment_variablesInsertAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        db_k = kwargs.get('k', '')
        db_v = kwargs.get('v', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO experiment_environment_variables (experimentid, k, v) VALUES (%d, '%s', '%s');""" % (int(db_expid), db_k, db_v))
        return
    
    def branches_completedSelectAllWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT experimentid, branchid, jobid, hostname, rundir FROM branches_completed WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_completedSelectCountWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT COUNT(branchid) FROM branches_completed WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_runningSelectBranchIDWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_running WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_runningSelectCountWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT COUNT(branchid) FROM branches_running WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_queuedSelectBranchIDWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_queued WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_queuedSelectBranchIDWhereCheckpointed(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_queued WHERE experimentid = %d AND branchid IN (SELECT branchid FROM jobs WHERE state = 'K');""" % (int(db_expid)))
    
    def branches_insignificantSelectBranchIDWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_insignificant WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_insignificantSelectBranchIDWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_insignificant WHERE branchid = %d;""" % (int(db_branchid)))
    
    def branches_persistently_failedSelectBranchIDWhere(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_persistently_failed WHERE experimentid = %d;""" % (int(db_expid)))
    
    def branches_persistently_failedSelectBranchIDWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT branchid FROM branches_persistently_failed WHERE branchid = %d;""" % (int(db_branchid)))
    
    def pending_directories_to_removeSelectAllAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT hostname, path FROM pending_directories_to_remove;""")
    
    def pending_directories_to_removeDeleteAllWhere(self, **kwargs):
        db_hostname = kwargs.get('hostname', '')
        db_jobdir = kwargs.get('jobdir', '')
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM pending_directories_to_remove WHERE hostname = '%s' AND path = '%s';""" % (db_hostname, db_jobdir))
    
    def pending_directories_to_removeDeleteAll(self):
        self._SqlDataProvider.ExecuteWriteQuery("""DELETE FROM pending_directories_to_remove;""")
    
    def pending_directories_to_removeInsertAllWhere(self, **kwargs):
        db_hostname = kwargs.get('hostname', '')
        db_path = kwargs.get('path', '')
        self._SqlDataProvider.ExecuteWriteQuery("""INSERT INTO pending_directories_to_remove (hostname, path) VALUES ('%s','%s');""" % (db_hostname, db_path))
        return
    
    def branch_scriptsSelectAllWhereSimname(self, **kwargs):
        db_simname = kwargs.get('simname', '')
        return self.ExecuteLegacyReadQuery("""SELECT simname, scriptname, filename, nargs FROM branch_scripts WHERE simname = '%s';""" % (db_simname))
    
    def branch_scriptsSelectAllWhereScriptname(self, **kwargs):
        db_scriptname = kwargs.get('scriptname', '')
        return self.ExecuteLegacyReadQuery("""SELECT simname, scriptname, filename, nargs FROM branch_scripts WHERE scriptname = '%s';""" % (db_scriptname))
    
    def jobs_branches_SelectCompletedJobsWhereExperimentByBranchID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT b.id, b.parentid, b.probability, b.branchdescription, b.submittime, j.starttime, j.endtime, j.id, j.hostname FROM branches b, jobs j WHERE b.experimentid = %d AND b.id = j.branchid AND j.state = 'C' ORDER BY b.id ASC;""" % (int(db_expid)))
    
    def jobs_branches_SelectCompletedJobsWhereExperimentByHostname(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT j.hostname, j.rundir FROM branches b, jobs j WHERE experimentid = %d AND b.id = j.branchid AND j.state = 'C' ORDER BY j.hostname;""" % (int(db_expid)))
    
    def jobs_branches_experimentsSelectRunningJobsWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT e.checkpointscript, j.id, j.rundir, j.hostname, j.pid FROM experiments e, jobs j, branches b WHERE e.id = b.experimentid AND j.branchid = b.id AND j.state = 'R' AND b.id = %d;""" % (int(db_branchid)))
    
    def jobs_branches_experimentsSelectRunningJobsAll(self):
        return self.ExecuteLegacyReadQuery("""SELECT e.checkpointscript, b.id, j.id, j.rundir, j.hostname, j.pid FROM experiments e, jobs j, branches b WHERE e.id = b.experimentid AND j.branchid = b.id AND j.state = 'R';""")
    
    def jobs_branches_experimentsSelectRunningJobsWhereExpID(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self.ExecuteLegacyReadQuery("""SELECT e.checkpointscript, b.id, j.id, j.rundir, j.hostname, j.pid FROM experiments e, jobs j, branches b WHERE e.id = b.experimentid AND j.branchid = b.id AND j.state = 'R' AND e.id = %d;""" % (int(db_expid)))
    
    def branches_branches_statesSelectJobsWhereBranchID(self, **kwargs):
        db_branchid = kwargs.get('branchid', '')
        return self.ExecuteLegacyReadQuery("""SELECT b.id, b.parentid, b.branchdescription, b.submittime, st.state, b.priority FROM branches b, branches_states st WHERE b.id = %d AND st.branchid=%d ORDER BY id;""" % (int(db_branchid), int(db_branchid)))
    
    def branchesSelectBranchIDWherePriorityMismatch(self):
        return self.ExecuteLegacyReadQuery("""SELECT b1.id, b2.id FROM branches b1, branches b2 WHERE b1.id <> b2.id AND b1.id IN (SELECT branchid FROM branches_running) AND b2.id IN (SELECT branchid FROM branches_queued) AND b2.experimentid NOT IN (SELECT id FROM experiments WHERE is_checkpointed='y') AND b1.priority < b2.priority;""")
    
    def branches_running_experimentsSelectExperimentIDWhereBranchesRunning(self):
        return self.ExecuteLegacyReadQuery("""SELECT DISTINCT br.experimentid FROM branches_running br, experiments e WHERE e.is_checkpointed = 'y' AND br.experimentid = e.id;""")
    
    def branches_experimentsSelectAllWhereBranchesRunning(self):
        return self.ExecuteLegacyReadQuery("""SELECT b.id, b.jobscript, b.experimentid, b.parentid, b.branchdescription, b.terminate_early, b.probability, b.probability_threshold_multiplier, e.probability_threshold, b.executable FROM branches b, experiments e WHERE b.experimentid = e.id AND e.is_checkpointed = 'n' AND b.id NOT IN (SELECT branchid FROM jobs WHERE state = 'R' OR state = 'C') AND b.id NOT IN (SELECT branchid FROM branches_persistently_failed) ORDER BY b.priority DESC, b.id;""")
    
    def branches_experimentsSelectAllWhereBranchesRunningLimited(self, **kwargs):
        db_limit = kwargs.get('limit', '')
        return self.ExecuteLegacyReadQuery("""SELECT b.id, b.jobscript, b.experimentid, b.parentid, b.branchdescription, b.terminate_early, b.probability, b.probability_threshold_multiplier, e.probability_threshold, b.executable FROM branches b, experiments e WHERE b.experimentid = e.id AND e.is_checkpointed = 'n' AND b.id NOT IN (SELECT branchid FROM jobs WHERE state = 'R' OR state = 'C') AND b.id NOT IN (SELECT branchid FROM branches_persistently_failed) ORDER BY b.priority DESC, b.id LIMIT %d;""" % (int(db_limit)))
    
    def branches_jobs_job_output_infoSelectAllWhereBranchesComplete(self, **kwargs):
        db_expid = kwargs.get('expid', '')
        return self._SqlDataProvider.ExecuteLegacyReadQuery("""SELECT b.id, b.parentid, j.id, ji.v FROM branches b, jobs j, job_output_info ji WHERE b.experimentid = %d AND b.id = j.branchid AND j.state = 'C' AND j.id = ji.jobid AND ji.k = 'sec' ORDER BY b.id;""" % (int(db_expid)))
    
    def ShowTables(self):
        # Despite the name, this returns both tables and views.
        return self.ExecuteLegacyReadQuery("""SHOW TABLES;""")
    
    def drop_database_tables(self):
        db = AdaptConfig.get_database_db()
        try:
            self._SqlDataProvider.ExecuteWriteQuery("""USE %s; SET foreign_key_checks = 0; DROP table experiments; DROP table branches; DROP table jobs; DROP table branch_input_blobs; DROP table branch_input_refs; DROP table job_output_blobs; DROP table job_output_info; DROP table job_checkpoints_taken; DROP table job_output_ls; DROP table experiment_environment_variables; DROP table control_messages; DROP table pending_directories_to_remove; DROP table simulators; DROP table simulators_prereqs; DROP table simulators_installer_prereqs; DROP table branch_scripts; DROP table dynamic_importance_measures; DROP view branches_aborts; DROP view branches_running; DROP view branches_insignificant; DROP view branches_queued; DROP view branches_completed; DROP view branches_checkpointed; DROP view branches_persistently_failed; DROP view branches_states; SET foreign_key_checks = 1;""" % (db))
        except:
            pass
    
    def dualSelectLastInsertID(self):
        return self.ExecuteLegacyReadQuery("""SELECT LAST_INSERT_ID() FROM dual;""")
    
    def MySQLShowProcesslist(self):
        return self.ExecuteLegacyReadQuery("""SHOW PROCESSLIST;""")
    
    def MySQLKillProcess(self, **kwargs):
        db_pid = kwargs.get('pid', '')
        self._SqlDataProvider.ExecuteWriteQuery("""KILL %d;""" % (int(db_pid)))
    
    def db_close_handles(self):
        global _db_cursor
        global _db_conn
        global _db_readcursor
        if _db_cursor:
            _db_cursor.close()
            if not (_db_readcursor is None):
                _db_readcursor.close()
                
            _db_conn.close()
            _db_conn = None
            _db_cursor = None
            _db_readcursor = None
    
    def db_lock(self):
        assert _db_conn
        assert _db_cursor
        uid = os.getuid()
    
    def db_unlock(self):
        assert _db_conn
        assert _db_cursor
        uid = os.getuid()
    
    def db_escape(self, val):
        if sql_connect_mode == SqlConnectMode.mysql:
            from mysql.connector import conversion
            return conversion.MySQLConverter().escape(val)
        else:
            raise RuntimeError("Unknown sql_connect_mode value.")
    
    def create_needed_database_tables(self):

        engine = 'InnoDB'
        db = AdaptConfig.get_database_db()
        ddl = """
    use %(database)s

    create table if not exists experiments(id int AUTO_INCREMENT, name varchar(255), description varchar(255), starttime datetime, probability_threshold double, checkpointscript longblob, localstoragedir varchar(255), is_checkpointed varchar(1), retries_permitted int, PRIMARY KEY(id)) %(tablesuffix)s

    create table if not exists branches(id int AUTO_INCREMENT, experimentid int, parentid int, branchdescription varchar(255), jobscript longblob, submittime datetime, probability double, probability_threshold_multiplier double, terminate_early int, priority int, executable varchar(255), PRIMARY KEY(id), INDEX(experimentid), FOREIGN KEY(experimentid) REFERENCES experiments(id) ON DELETE cascade, INDEX(parentid), FOREIGN KEY(parentid) REFERENCES branches(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists jobs(id int AUTO_INCREMENT, branchid int, state char(1), hostname varchar(255), rundir varchar(255), starttime datetime, endtime datetime, pid varchar(255), PRIMARY KEY(id), INDEX(branchid), FOREIGN KEY(branchid) REFERENCES branches(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists branch_input_blobs(branchid int, filename varchar(255), permissions int, dat longblob, INDEX(branchid), FOREIGN KEY(branchid) REFERENCES branches(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists branch_input_refs(branchid int, hostname varchar(255), filenamehere varchar(255), filenamethere varchar(255), INDEX(branchid), FOREIGN KEY(branchid) REFERENCES branches(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists job_output_blobs(jobid int, filename varchar(255), permissions int, dat longblob, INDEX(jobid), FOREIGN KEY(jobid) REFERENCES jobs(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists job_output_info(jobid int, k varchar(255), v varchar(255), INDEX(jobid), FOREIGN KEY(jobid) REFERENCES jobs(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists job_checkpoints_taken(jobid int, checkpoint_time datetime, resume_time datetime, INDEX(jobid), FOREIGN KEY(jobid) REFERENCES jobs(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists job_output_ls(jobid int, path varchar(255), size integer, INDEX(jobid), FOREIGN KEY(jobid) REFERENCES jobs(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists experiment_environment_variables(experimentid int, k varchar(255), v varchar(255), INDEX(experimentid), FOREIGN KEY(experimentid) REFERENCES experiments(id) ON DELETE cascade) %(tablesuffix)s

    create table if not exists control_messages(k varchar(255), v varchar(255)) %(tablesuffix)s

    create table if not exists pending_directories_to_remove(hostname varchar(255), path varchar(255))

    create table if not exists simulators(simname varchar(255), webwrapperscript longblob, mainwrapperscript longblob, primary key(simname)) %(tablesuffix)s

    create table if not exists simulators_prereqs(simname varchar(255), filename varchar(255), inputtype char(64), INDEX(simname), FOREIGN KEY(simname) references simulators(simname)) %(tablesuffix)s

    create table if not exists simulators_installer_prereqs(simname varchar(255), filename varchar(255), payload longblob) %(tablesuffix)s

    create table if not exists branch_scripts(simname varchar(255), scriptname varchar(255), filename varchar(255), nargs int)
    
    create table if not exists dynamic_importance_measures(expid int, code varchar(255), variable varchar(255), name varchar(255), type int, consequence_file varchar(255), value_nonoccurrence varchar(255), value double)

    create table if not exists branching_condition_names(expid int, code varchar(255), name varchar(255))

    create or replace view branches_aborts as (select branchid, count(branchid) as aborts from jobs where state = 'A' group by branchid)

    create or replace view branches_running as select e.id as experimentid, b.id as branchid from branches b, experiments e where b.experimentid = e.id and b.id in (select branchid from jobs where state = 'R')

    create or replace view branches_insignificant as select e.id as experimentid, b.id as branchid from branches b, experiments e where b.experimentid = e.id and b.id not in (select branchid from jobs where state = 'R' or state = 'C' or state = 'K') and b.probability < (e.probability_threshold * b.probability_threshold_multiplier)

    create or replace view branches_queued as select e.id as experimentid, b.id as branchid from branches b, experiments e where b.experimentid = e.id and b.id not in (select branchid from jobs where state = 'C' or state = 'R') and b.id not in (select branchid from branches_insignificant)

    create or replace view branches_completed as select e.id as experimentid, b.id as branchid, j.id as jobid, j.hostname as hostname, j.rundir as rundir from branches b, experiments e, jobs j where b.experimentid = e.id and b.id = j.branchid and j.state = 'C'

    create or replace view branches_checkpointed as select e.id as experimentid, b.id as branchid from branches b, experiments e where b.experimentid = e.id and b.id not in (select branchid from jobs where state = 'C' or state = 'R') and b.id in (select branchid from jobs where state = 'K')

    create or replace view branches_persistently_failed as (select e.id as experimentid, b.id as branchid from experiments e, branches b where e.id = b.experimentid and b.id in (select branchid from branches_aborts where aborts >= e.retries_permitted and branchid not in (select branchid from branches_completed)))

    create or replace view branches_states as (select branchid, 'C' as state from branches_completed) union (select branchid, 'R' as state from branches_running) union (select branchid, 'Q' as state from branches_queued) order by branchid
    """ % {'tablesuffix': 'engine=InnoDB', 'database': db}
        ddl_lines = ddl.split('\n')
        ddl_lines = [x.strip() for x in ddl_lines]
        ddl_lines = [x for x in ddl_lines if x != '']
        
        for statement in ddl_lines:
            if not re.search("^#", statement):
                try:
                    self._SqlDataProvider.ExecuteWriteQuery(statement)
                except Exception:
                    traceback.print_exc()
                    sys.exit(1)
