#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ##########################
# ADAPT copyright statement:
# Copyright 2006 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software. This software is distributed under the GNU Lesser General Public License version 3 or later.
# ##########################
# ADAPT module description for bashselect.py: This module does some command formatting for ADAPT. It is used by numerous other ADAPT scripts.
from __future__ import print_function
import os
import code


def bashselect(items):
    if not isinstance(items, list):
        raise ValueError('input is not a list')
    if len(items) == 0:
        raise ValueError('empty list given')
    fmtlen = len(repr(len(items)))
    cols = int(os.getenv('COLUMNS') or 80) - 1
    rows = int(os.getenv('LINES') or 24) - 2
    labellen = fmtlen + 2
    maxw = -1
    for i in items:
        if len(i) > maxw:
            maxw = len(i)
    col2 = 0
    if (labellen + maxw + 2 + labellen + maxw) <= cols and len(items) > rows:
        col2 = 1
    n = 1
    if col2:
        i1 = items[0:1 + len(items) / 2]
        i2 = items[1 + len(items) / 2:]
        i1_origlen = len(i1)
        while len(i1) > 0:
            if len(i2) > 0:
                print('%*d) %*s  %*d) %*s' % (fmtlen, n, maxw, i1[0],
                                              fmtlen, n + i1_origlen, maxw, i2[0]))
                i1.pop(0)
                i2.pop(0)
            else:
                print('%*d) %*s' % (fmtlen, n, maxw, i1[0]))
                assert len(i1) == 1
                i1.pop(0)
            n += 1
    else:
        for i in items:
            print('%*d) %s' % (fmtlen, n, i))
            n += 1
    print()
    
    out = None
    c = code.InteractiveConsole()
    while True:
        prompt = 'Enter a choice, or x to abort: '
        try:
            response = c.raw_input(prompt).strip()
        except EOFError:
            break
        if response == 'x':
            break
        elif not response.isdigit():
            pass
        else:
            val = int(response)
            if val >= 1 and val <= len(items):
                val -= 1
                out = (val, items[val])
                break
    return out

if __name__ == "__main__": # when run as a script
    print(bashselect([chr(x) * 2 for x in range(97, 122)]))
    print(bashselect([chr(x) * 34 for x in range(97, 122)]))
    print(bashselect([chr(x) * 35 for x in range(97, 122)]))
